import { objectWithoutProperties as _objectWithoutProperties, asyncToGenerator as _asyncToGenerator } from './_virtual/_rollupPluginBabelHelpers.js';
import { EventEmitter } from 'events';
import { resolve, join } from 'path';
import { concatAST, parse, Source } from 'graphql';
import chalk from 'chalk';
import { mkdirp, writeFile, readFile } from 'fs-extra';
import { loadConfigSync } from 'graphql-config';
import { getGraphQLProjectForSchemaPath, resolvePathRelativeToConfig, getGraphQLSchemaPaths, getGraphQLProjects } from 'graphql-config-utilities';
import { compile, isOperation } from 'graphql-tool-utilities';
export { EnumFormat, ExportFormat } from './types.mjs';
import { generateSchemaTypes } from './print/schema/index.mjs';
import { printDocument } from './print/document/index.mjs';
export { AbstractGraphQLFilesystem } from './filesystem/graphql-filesystem.mjs';

const _excluded = ["cwd"];
class Builder extends EventEmitter {
  // workspace graphql configuration
  // see: https://github.com/prisma/graphql-config
  // projectName -> {filePath -> document}
  // NOTE: projectName can be undefined for nameless graphql-config projects
  constructor(_ref) {
    var _this;

    let {
      cwd
    } = _ref,
        options = _objectWithoutProperties(_ref, _excluded);

    super();
    _this = this;
    this.options = void 0;
    this.config = void 0;
    this.documentMapByProject = new Map();
    this.filesystem = void 0;

    this.handleDocumentUpdate = /*#__PURE__*/function () {
      var _ref2 = _asyncToGenerator(function* (filePath, projectConfig) {
        try {
          yield _this.updateDocumentForFile(filePath, projectConfig);
        } catch (error) {
          _this.emit('error', error);

          return;
        }

        yield _this.generateDocumentTypes();
      });

      return function (_x, _x2) {
        return _ref2.apply(this, arguments);
      };
    }();

    this.handleDocumentDelete = /*#__PURE__*/function () {
      var _ref3 = _asyncToGenerator(function* (filePath, projectConfig) {
        const documents = _this.documentMapByProject.get(projectConfig.name);

        if (documents) {
          documents.delete(filePath);
        }

        yield _this.generateDocumentTypes();
      });

      return function (_x3, _x4) {
        return _ref3.apply(this, arguments);
      };
    }();

    this.handleSchemaUpdate = /*#__PURE__*/function () {
      var _ref4 = _asyncToGenerator(function* (schemaPath) {
        try {
          _this.emit('start:schema');

          yield _this.generateSchemaTypes(schemaPath);

          _this.emit('end:schema');

          yield _this.generateDocumentTypes();
        } catch (error) {// intentional noop
        }
      });

      return function (_x5) {
        return _ref4.apply(this, arguments);
      };
    }();

    this.options = options;
    const configPath = cwd ? resolve(cwd) : undefined;
    this.config = options.config ? options.config : loadConfigSync({
      rootDir: configPath
    });
  }

  once(event, handler) {
    return super.once(event, handler);
  }

  on(event, handler) {
    return super.on(event, handler);
  }

  emit(event, ...args) {
    return super.emit(event, ...args);
  }

  run({
    watch: watchGlobs = false,
    graphQLFilesystem
  } = {}) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      let schemaPaths;

      try {
        schemaPaths = getGraphQLSchemaPaths(_this2.config);
      } catch (error) {
        _this2.emit('error', error);

        return;
      }

      const filesystem = graphQLFilesystem !== null && graphQLFilesystem !== void 0 ? graphQLFilesystem : yield defaultFilesytem();
      _this2.filesystem = filesystem;

      if (watchGlobs) {
        filesystem.on('change:document', _this2.handleDocumentUpdate).on('delete:document', _this2.handleDocumentDelete).on('change:schema', _this2.handleSchemaUpdate); // wait for all watchers to be ready

        yield filesystem.watch(_this2.config);
      }

      try {
        _this2.emit('start:schema');

        yield Promise.all(schemaPaths.map(schemaPath => _this2.generateSchemaTypes(schemaPath)));

        _this2.emit('end:schema');
      } catch (error) {
        _this2.emit('error', error);

        return;
      }

      try {
        yield Promise.all(getGraphQLProjects(_this2.config).map(projectConfig => _this2.updateDocumentsForProject(filesystem, projectConfig)));
      } catch (error) {
        _this2.emit('error', error);

        return;
      }

      yield _this2.generateDocumentTypes();
    })();
  }

  stop() {
    if (this.filesystem && typeof this.filesystem.dispose === 'function') {
      this.filesystem.dispose();
    }
  }

  generateSchemaTypes(schemaPath) {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      const projectConfig = getGraphQLProjectForSchemaPath(_this3.config, schemaPath);
      const schemaTypesPath = getSchemaTypesPath(projectConfig, _this3.options);
      const definitions = generateSchemaTypes(yield projectConfig.getSchema(), _this3.options);
      yield mkdirp(schemaTypesPath);
      yield Promise.all(Array.from(definitions.entries()).map(([fileName, definition]) => writeFile(join(schemaTypesPath, fileName), definition)));

      _this3.emit('build:schema', {
        schemaPath,
        schemaTypesPath
      });
    })();
  }

  generateDocumentTypes() {
    var _this4 = this;

    return _asyncToGenerator(function* () {
      _this4.emit('start:docs');

      _this4.checkForDuplicateOperations();

      _this4.checkForDuplicateFragments();

      yield Promise.all(Array.from(_this4.documentMapByProject.entries()).map(([projectName, documents]) => _this4.generateDocumentTypesForProject(_this4.config.getProject(projectName), documents)));

      _this4.emit('end:docs');
    })();
  }

  checkForDuplicateOperations() {
    getDuplicateOperations(this.documentMapByProject).forEach(({
      projectName,
      duplicates
    }) => {
      if (duplicates.length) {
        duplicates.forEach(({
          operationName,
          filePaths
        }) => {
          const message = `GraphQL operations must have a unique name. The operation ${chalk.bold(operationName)} is declared in:\n ${filePaths.sort().join('\n ')}${projectName ? ` (${chalk.bold(projectName)})` : ''}`;
          this.emit('error', new Error(message));
        });
      }
    });
  }

  checkForDuplicateFragments() {
    getDuplicateFragments(this.documentMapByProject).forEach(({
      projectName,
      duplicates
    }) => {
      if (duplicates.length) {
        duplicates.forEach(({
          fragmentName,
          filePaths
        }) => {
          const message = `GraphQL fragments must have a unique name. The fragment ${chalk.bold(fragmentName)} is declared in:\n ${filePaths.sort().join('\n ')}${projectName ? ` (${chalk.bold(projectName)})` : ''}`;
          this.emit('error', new Error(message));
        });
      }
    });
  }

  generateDocumentTypesForProject(projectConfig, documents) {
    var _this5 = this;

    return _asyncToGenerator(function* () {
      let ast;

      try {
        ast = compile(yield projectConfig.getSchema(), concatAST(Array.from(documents.values())));
      } catch (error) {
        _this5.emit('error', error);

        return;
      }

      try {
        yield Promise.all(Array.from(groupOperationsAndFragmentsByFile(ast).values()).map(file => _this5.writeDocumentFile(file, ast, projectConfig)));
      } catch (error) {// intentional noop
      }
    })();
  }

  writeDocumentFile(file, ast, projectConfig) {
    var _this6 = this;

    return _asyncToGenerator(function* () {
      const definitionPath = `${file.path}.d.ts`;
      yield writeFile(definitionPath, _this6.getDocumentDefinition(file, ast, projectConfig));

      _this6.emit('build:docs', {
        documentPath: file.path,
        definitionPath,
        operation: file.operation,
        fragments: file.fragments
      });
    })();
  }

  getDocumentDefinition(file, ast, projectConfig) {
    try {
      return printDocument(file, ast, {
        enumFormat: this.options.enumFormat,
        exportFormat: this.options.exportFormat,
        addTypename: this.options.addTypename,
        schemaTypesPath: getSchemaTypesPath(projectConfig, this.options)
      });
    } catch ({
      message
    }) {
      const error = new Error(`Error in ${file.path}: ${message[0].toLowerCase()}${message.slice(1)}`);
      this.emit('error', error);
      throw error;
    }
  }

  updateDocumentsForProject(filesystem, projectConfig) {
    var _this7 = this;

    return _asyncToGenerator(function* () {
      const filePaths = yield filesystem.getGraphQLProjectIncludedFilePaths(projectConfig);
      return Promise.all(filePaths.map(filePath => _this7.updateDocumentForFile(filePath, projectConfig)));
    })();
  }

  updateDocumentForFile(filePath, projectConfig) {
    var _this8 = this;

    return _asyncToGenerator(function* () {
      const contents = yield readFile(filePath, 'utf8');
      return _this8.setDocumentForFilePath(filePath, projectConfig, contents);
    })();
  }

  setDocumentForFilePath(filePath, projectConfig, contents) {
    let documents = this.documentMapByProject.get(projectConfig.name);

    if (!documents) {
      documents = new Map();
      this.documentMapByProject.set(projectConfig.name, documents);
    }

    if (contents.trim().length === 0) {
      return undefined;
    }

    const document = parse(new Source(contents, filePath));
    documents.set(filePath, document);
    return document;
  }

}

function getSchemaTypesPath(projectConfig, options) {
  if (typeof projectConfig.extensions.schemaTypesPath === 'string') {
    return resolvePathRelativeToConfig(projectConfig, projectConfig.extensions.schemaTypesPath);
  }

  return resolvePathRelativeToConfig(projectConfig, join(options.schemaTypesPath, `${projectConfig.name ? `${projectConfig.name}-` : ''}types`));
}

function defaultFilesytem() {
  return _defaultFilesytem.apply(this, arguments);
}

function _defaultFilesytem() {
  _defaultFilesytem = _asyncToGenerator(function* () {
    const {
      DefaultGraphQLFilesystem
    } = yield import('./filesystem/default-graphql-filesystem.mjs');
    return new DefaultGraphQLFilesystem();
  });
  return _defaultFilesytem.apply(this, arguments);
}

function groupOperationsAndFragmentsByFile({
  operations,
  fragments
}) {
  return Object.values(operations).concat(Object.values(fragments)).reduce((map, item) => {
    if (!item.filePath) {
      return map;
    }

    let file = map.get(item.filePath);

    if (!file) {
      file = {
        path: item.filePath,
        operation: undefined,
        fragments: []
      };
      map.set(item.filePath, file);
    }

    if (isOperation(item)) {
      file.operation = item;
    } else {
      file.fragments.push(item);
    }

    return map;
  }, new Map());
}

function getDuplicateOperations(documentsMapByProject) {
  return Array.from(documentsMapByProject.entries()).map(([projectName, documents]) => {
    return {
      projectName,
      duplicates: getDuplicateProjectOperations(documents)
    };
  });
}

function getDuplicateFragments(documentsMapByProject) {
  return Array.from(documentsMapByProject.entries()).map(([projectName, documents]) => {
    return {
      projectName,
      duplicates: getDuplicateProjectFragments(documents)
    };
  });
}

function getDuplicateProjectOperations(documents) {
  const operations = new Map();
  Array.from(documents.entries()).forEach(([filePath, document]) => {
    document.definitions.filter(isOperationDefinition).forEach(definition => {
      const {
        name
      } = definition;

      if (name && name.value) {
        const map = operations.get(name.value);

        if (map) {
          map.add(filePath);
        } else {
          operations.set(name.value, new Set([filePath]));
        }
      }
    });
  });
  return Array.from(operations.entries()).filter(([, filePaths]) => filePaths.size > 1).map(([operationName, filePath]) => {
    return {
      operationName,
      filePaths: Array.from(filePath)
    };
  });
}

function getDuplicateProjectFragments(documents) {
  const fragments = new Map();
  Array.from(documents.entries()).forEach(([filePath, document]) => {
    document.definitions.filter(isFragmentDefinition).forEach(definition => {
      const {
        name
      } = definition;

      if (name && name.value) {
        const map = fragments.get(name.value);

        if (map) {
          map.add(filePath);
        } else {
          fragments.set(name.value, new Set([filePath]));
        }
      }
    });
  });
  return Array.from(fragments.entries()).filter(([, filePaths]) => filePaths.size > 1).map(([fragmentName, filePath]) => {
    return {
      fragmentName,
      filePaths: Array.from(filePath)
    };
  });
}

function isOperationDefinition(definition) {
  return definition.kind === 'OperationDefinition';
}

function isFragmentDefinition(definition) {
  return definition.kind === 'FragmentDefinition';
}

export { Builder };
